@@ -26,7 +26,7 @@ contact."""
 
 import msnp
 import p2p
-from switchboard_manager import SwitchboardClient
+from switchboard_manager import SwitchboardHandler
 from papyon.event import EventsDispatcher
 from papyon.profile import NetworkID
 
@@ -74,6 +74,10 @@ class ConversationInterface(object):
     """Interface implemented by all the Conversation objects, a Conversation
     object allows the user to communicate with one or more peers"""
 
+    @property
+    def max_message_size(self):
+        raise NotImplementedError
+
     def send_text_message(self, message):
         """Send a message to all persons in this conversation.
 
@@ -135,6 +139,17 @@ class ConversationMessage(object):
         self.formatting = formatting
         self.msn_objects = msn_objects
 
+    def split(self, size):
+        """Split message in parts of given size."""
+        #TODO make sure we don't cut a smiley alias
+        parts = []
+        cs = [self.content[i:i+size] for i in range(0, len(self.content), size)]
+        for c in cs:
+            part = ConversationMessage(c, self.formatting, self.msn_objects)
+            part.display_name = self.display_name
+            parts.append(part)
+        return parts
+
 class TextFormat(object):
 
     DEFAULT_FONT = 'MS Sans Serif'
@@ -227,6 +242,8 @@ class TextFormat(object):
 
     def __parse(self, format):
         for property in format.split(';'):
+            if not property:
+                continue
             key, value =  [p.strip(' \t|').upper() \
                     for p in property.split('=', 1)]
             if key == 'FN':
@@ -288,7 +305,18 @@ class AbstractConversation(ConversationInterface, EventsDispatcher):
 
         self.__last_received_msn_objects = {}
 
-    def send_text_message(self, message):
+    @property
+    def max_message_size(self):
+        return 1500
+
+    def send_text_message(self, message, callback=None, errback=None):
+        if len(message.content) > self.max_message_size:
+            logger.info("Message content is too large, message will be split")
+            parts = message.split(self.max_message_size)
+            for part in parts:
+                self.send_text_message(part)
+            return
+
         if len(message.msn_objects) > 0:
             body = []
             for alias, msn_object in message.msn_objects.iteritems():
@@ -297,29 +325,29 @@ class AbstractConversation(ConversationInterface, EventsDispatcher):
                 body.append(str(msn_object))
                 # FIXME : we need to distinguish animemoticon and emoticons
                 # and send the related msn objects in separated messages
-            self._send_message(("text/x-mms-animemoticon",), '\t'.join(body))
+            self._send_message_ex(("text/x-mms-animemoticon",), '\t'.join(body))
 
         content_type = ("text/plain","utf-8")
         body = message.content.encode("utf-8")
-        ack = msnp.MessageAcknowledgement.HALF
         headers = {}
         if message.formatting is not None:
             headers["X-MMS-IM-Format"] = str(message.formatting)
 
-        self._send_message(content_type, body, headers, ack)
+        callback = (self._on_text_message_sent, message)
+        self._send_message_ex(content_type, body, headers, callback)
 
     def send_nudge(self):
         content_type = "text/x-msnmsgr-datacast"
         body = "ID: 1\r\n\r\n".encode('UTF-8') #FIXME: we need to figure out the datacast objects :D
-        ack = msnp.MessageAcknowledgement.NONE
-        self._send_message(content_type, body, ack=ack)
+
+        callback = (self._on_nudge_sent,)
+        self._send_message_ex(content_type, body, {}, callback)
 
     def send_typing_notification(self):
         content_type = "text/x-msmsgscontrol"
         body = "\r\n\r\n".encode('UTF-8')
-        headers = { "TypingUser" : self._client.profile.account.encode('UTF_8') }
-        ack = msnp.MessageAcknowledgement.NONE
-        self._send_message(content_type, body, headers, ack)
+        headers = { "TypingUser" : self._client.profile.account }
+        self._send_message_ex(content_type, body, headers)
 
     def invite_user(self, contact):
         raise NotImplementedError
@@ -327,8 +355,16 @@ class AbstractConversation(ConversationInterface, EventsDispatcher):
     def leave(self):
         raise NotImplementedError
 
-    def _send_message(self, content_type, body, headers={},
-            ack=msnp.MessageAcknowledgement.HALF):
+    def _send_message_ex(self, content_type, body, headers={},
+            callback=None, errback=None):
+        message = msnp.Message(self._client.profile)
+        for key, value in headers.iteritems():
+            message.add_header(key, value)
+        message.content_type = content_type
+        message.body = body
+        return self._send_message(message, callback, errback)
+
+    def _send_message(self, message, callback=None, errback=None):
         raise NotImplementedError
 
     def _on_contact_joined(self, contact):
@@ -366,17 +402,23 @@ class AbstractConversation(ConversationInterface, EventsDispatcher):
             msn_objects = {}
             parts = message.body.split('\t')
             logger.debug(parts)
-            for i in [i for i in range(len(parts)) if not i % 2]:
+            for i in range(0, len(parts), 2):
                 if parts[i] == '': break
-                msn_objects[parts[i]] = p2p.MSNObject.parse(self._client,
-                        parts[i+1])
+                try:
+                    msn_object = p2p.MSNObject.parse(self._client, parts[i+1])
+                except:
+                    msn_object = None
+                msn_objects[parts[i]] = msn_object
             self.__last_received_msn_objects = msn_objects
         elif message_type == 'text/x-msnmsgr-datacast' and \
                 message.body.strip() == "ID: 1":
             self._dispatch("on_conversation_nudge_received", sender)
 
-    def _on_message_sent(self, message):
-        pass
+    def _on_nudge_sent(self):
+        self._dispatch("on_conversation_nudge_sent")
+
+    def _on_text_message_sent(self, message):
+        self._dispatch("on_conversation_message_sent", message)
 
     def _on_error(self, error_type, error):
         self._dispatch("on_conversation_error", error_type, error)
@@ -401,47 +443,58 @@ class ExternalNetworkConversation(AbstractConversation):
     def leave(self):
         self._client._unregister_external_conversation(self)
 
-    def _send_message(self, content_type, body, headers={},
-            ack=msnp.MessageAcknowledgement.HALF):
-        if content_type[0]  in ['text/x-mms-emoticon',
-                                'text/x-mms-animemoticon']:
-            return
-        message = msnp.Message(self._client.profile)
-        for key, value in headers.iteritems():
-            message.add_header(key, value)
-        message.content_type = content_type
-        message.body = body
+    def _send_message(self, message, callback=None, errback=None):
+        content_type = message.content_type[0]
+        if content_type in ['text/x-mms-emoticon', 'text/x-mms-animemoticon']:
+            return # don't send icons to external contacts
         for contact in self.participants:
             self._client._protocol.\
-                    send_unmanaged_message(contact, message)
+                    send_unmanaged_message(contact, message, callback, errback)
 
 
-class SwitchboardConversation(AbstractConversation, SwitchboardClient):
+class SwitchboardConversation(AbstractConversation, SwitchboardHandler):
     def __init__(self, client, contacts):
-        SwitchboardClient.__init__(self, client, contacts, priority=0)
+        SwitchboardHandler.__init__(self, client, contacts, priority=0)
         AbstractConversation.__init__(self, client)
 
     @staticmethod
-    def _can_handle_message(message, switchboard_client=None):
+    def _can_handle_message(message, switchboard_handler=None):
         content_type = message.content_type[0]
-        if switchboard_client is None:
-            return content_type in ('text/plain', 'text/x-msnmsgr-datacast')
-        # FIXME : we need to not filter those 'text/x-mms-emoticon', 'text/x-mms-animemoticon'
+        if not switchboard_handler and content_type in ('text/x-msmsgscontrol'):
+            return False # don't create conversation on typing notification
         return content_type in ('text/plain', 'text/x-msmsgscontrol',
                 'text/x-msnmsgr-datacast', 'text/x-mms-emoticon',
                 'text/x-mms-animemoticon')
 
+    @staticmethod
+    def handle_message(client, message):
+        return SwitchboardConversation(client, ())
+
     def invite_user(self, contact):
         """Request a contact to join in the conversation.
 
             @param contact: the contact to invite.
             @type contact: L{profile.Contact}"""
-        SwitchboardClient._invite_user(self, contact)
+        SwitchboardHandler._invite_user(self, contact)
 
     def leave(self):
         """Leave the conversation."""
-        SwitchboardClient._leave(self)
+        SwitchboardHandler._leave(self)
 
-    def _send_message(self, content_type, body, headers={},
-            ack=msnp.MessageAcknowledgement.HALF):
-        SwitchboardClient._send_message(self, content_type, body, headers, ack)
+    def _send_message(self, message, callback=None, errback=None):
+        content_type = message.content_type[0]
+        ack = msnp.MessageAcknowledgement.HALF
+        if content_type in ("text/x-msnmsgr-datacast", "text/x-msmsgscontrol"):
+            ack = msnp.MessageAcknowledgement.NONE
+        SwitchboardHandler._send_message(self, message, ack,
+                callback=callback, errback=errback)
+
+    def _on_closed(self):
+        self._dispatch("on_conversation_closed")
+
+    def _on_switchboard_closed(self):
+        pass
+
+    def __repr__(self):
+        participants = ",".join(map(lambda p: p.account, self.total_participants))
+        return '<SwitchboardConversation participants="%s">' % participants
