@@ -23,15 +23,19 @@
 """Notification protocol Implementation
 Implements the protocol used to communicate with the Notification Server."""
 
-from base import BaseProtocol, ProtocolState
+from base import BaseProtocol
 from message import Message
-from constants import ProtocolConstant
+from constants import ProtocolConstant, ProtocolError, ProtocolState
 from challenge import _msn_challenge
 
 import papyon
 from papyon.gnet.message.HTTP import HTTPMessage
+from papyon.util.async import run
 from papyon.util.queue import PriorityQueue, LastElementQueue
 from papyon.util.decorator import throttled
+from papyon.util.encoding import decode_rfc2047_string
+from papyon.util.parsing import build_account, parse_account
+from papyon.util.timer import Timer
 import papyon.util.element_tree as ElementTree
 import papyon.profile as profile
 import papyon.service.SingleSignOn as SSO
@@ -41,6 +45,7 @@ import papyon.service.OfflineIM as OIM
 import hashlib
 import time
 import logging
+import uuid
 import urllib
 import gobject
 import xml.sax.saxutils as xml_utils
@@ -50,7 +55,7 @@ __all__ = ['NotificationProtocol']
 logger = logging.getLogger('papyon.protocol.notification')
 
 
-class NotificationProtocol(BaseProtocol, gobject.GObject):
+class NotificationProtocol(BaseProtocol, Timer):
     """Protocol used to communicate with the Notification Server
 
         @undocumented: do_get_property, do_set_property
@@ -58,23 +63,11 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
 
         @ivar state: the current protocol state
         @type state: integer
-        @see L{base.ProtocolState}"""
+        @see L{constants.ProtocolState}"""
     __gsignals__ = {
-            "authentication-failed" : (gobject.SIGNAL_RUN_FIRST,
-                gobject.TYPE_NONE,
-                ()),
-
-            "disconnected-by-other" : (gobject.SIGNAL_RUN_FIRST,
-                gobject.TYPE_NONE,
-                ()),
-
-            "server-down" : (gobject.SIGNAL_RUN_FIRST,
-                gobject.TYPE_NONE,
-                ()),
-
             "buddy-notification-received" : (gobject.SIGNAL_RUN_FIRST,
                 gobject.TYPE_NONE,
-                (object, object,)),
+                (object, object, object, object)),
 
             "mail-received" : (gobject.SIGNAL_RUN_FIRST,
                 gobject.TYPE_NONE,
@@ -89,14 +82,6 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
                 (object, object)),
             }
 
-    __gproperties__ = {
-            "state":  (gobject.TYPE_INT,
-                "State",
-                "The state of the communication with the server.",
-                0, 6, ProtocolState.CLOSED,
-                gobject.PARAM_READABLE)
-            }
-
     def __init__(self, client, transport, proxies={}, version=15):
         """Initializer
 
@@ -111,10 +96,10 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
             @type proxies: {type: string, proxy:L{gnet.proxy.ProxyInfos}}
         """
         BaseProtocol.__init__(self, client, transport, proxies)
-        gobject.GObject.__init__(self)
+        Timer.__init__(self)
         self.__state = ProtocolState.CLOSED
         self._protocol_version = version
-        self._url_callbacks = {} # tr_id=>callback
+        self._callbacks = {} # tr_id=>(callback, errback)
 
     # Properties ------------------------------------------------------------
     def __get_state(self):
@@ -135,7 +120,7 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         raise AttributeError, "unknown property %s" % pspec.name
 
     # Public API -------------------------------------------------------------
-    @throttled(2000, LastElementQueue())
+    @throttled(2, LastElementQueue())
     def set_presence(self, presence, client_id=0, msn_object=None):
         """Publish the new user presence.
 
@@ -152,7 +137,7 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
             self._send_command('CHG',
                     (presence, str(client_id), urllib.quote(str(msn_object))))
 
-    @throttled(2000, LastElementQueue())
+    @throttled(2, LastElementQueue())
     def set_display_name(self, display_name):
         """Sets the new display name
 
@@ -161,7 +146,7 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         self._send_command('PRP',
                 ('MFN', urllib.quote(display_name)))
 
-    @throttled(2000, LastElementQueue())
+    @throttled(2, LastElementQueue())
     def set_personal_message(self, personal_message='', current_media=None,
             signature_sound=None):
         """Sets the new personal message
@@ -183,7 +168,7 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
                 '<PSM>%s</PSM>'\
                 '<CurrentMedia>%s</CurrentMedia>'\
                 '<MachineGuid>%s</MachineGuid>'\
-            '</Data>' % (message, cm, self._client.machine_guid.upper())
+            '</Data>' % (message, cm, str(self._client.machine_guid).upper())
         self._send_command('UUX', payload=pm)
         self._client.profile._server_property_changed("personal-message",
                 personal_message)
@@ -191,7 +176,7 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
             self._client.profile._server_property_changed("current-media",
                 current_media)
 
-    @throttled(2000, LastElementQueue())
+    @throttled(2, LastElementQueue())
     def set_end_point_name(self, name="Papyon", idle=False):
         ep = '<EndpointData>'\
                 '<Capabilities>%s</Capabilities>'\
@@ -208,14 +193,19 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         self._send_command('UUX', payload=ep)
         self._send_command('UUX', payload=pep)
 
+    @throttled(2, LastElementQueue())
+    def set_privacy(self, privacy=profile.Privacy.BLOCK):
+        self._send_command("BLP", (privacy,))
+
     def signoff(self):
         """Logout from the server"""
-        self._send_command('OUT')
+        if self._state >= ProtocolState.AUTHENTICATING:
+            self._send_command('OUT')
         self._transport.lose_connection()
 
-    @throttled(7600, list())
-    def request_switchboard(self, priority, callback, *callback_args):
-        self.__switchboard_callbacks.add((callback, callback_args), priority)
+    @throttled(7, list())
+    def request_switchboard(self, priority, callback):
+        self.__switchboard_callbacks.add(callback, priority)
         self._send_command('XFR', ('SB',))
 
     def add_contact_to_membership(self, account,
@@ -270,10 +260,15 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
                     (domain, user, membership, network_id)
             self._send_command("RML", payload=payload)
 
-    def send_user_notification(self, message, contact, type):
-        self._send_command("UUN", (contact.account, type), message)
+    def send_user_notification(self, message, contact, contact_guid, type,
+            callback=None, errback=None):
+        account = build_account(contact, contact_guid)
+        arguments = (account, type)
+        tr_id = self._send_command("UUN", arguments, message, True)
+        self._callbacks[tr_id] = (callback, errback)
 
-    def send_unmanaged_message(self, contact, message):
+    def send_unmanaged_message(self, contact, message, callback=None,
+            errback=None):
         content_type = message.content_type[0]
         if content_type == 'text/x-msnmsgr-datacast':
             message_type = 3
@@ -281,16 +276,27 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
             message_type = 2
         else:
             message_type = 1
-        self._send_command('UUM',
+        tr_id = self._send_command('UUM',
                 (contact.account, contact.network_id, message_type),
-                payload=message)
+                payload=message, callback=callback)
+        self._callbacks[tr_id] = (None, errback)
 
     def send_url_request(self, url_command_args, callback):
         tr_id = self._send_command('URL', url_command_args)
-        self._url_callbacks[tr_id] = callback
+        self._callbacks[tr_id] = (callback, None)
 
-    def _parse_account(self, command, idx=0):
-        if self._protocol_version >= 18:
+    # Helpers ----------------------------------------------------------------
+    def __search_account(self, account, network_id=profile.NetworkID.MSN):
+        contact = self._client.address_book.search_contact(account, network_id)
+
+        if contact is None:
+            logger.warning("Contact (network_id=%d) %s not found" % \
+                    (network_id, account))
+
+        return contact
+
+    def __parse_network_and_account(self, command, idx=0):
+        if self._protocol_version >= 18 and ":" in command.arguments[idx]:
             temp = command.arguments[idx].split(":")
             network_id = int(temp[0])
             account = temp[1]
@@ -301,6 +307,12 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         idx += 1
         return idx, network_id, account
 
+    def __find_node(self, parent, name, default):
+        node = parent.find(name)
+        if node is not None and node.text is not None:
+            return node.text.encode("utf-8")
+        else:
+            return default
 
     # Handlers ---------------------------------------------------------------
     # --------- Connection ---------------------------------------------------
@@ -332,8 +344,8 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
                 host = command.arguments[1]
                 port = self._transport.server[1]
             session_id = command.arguments[3]
-            callback, callback_args = self.__switchboard_callbacks.pop(0)
-            callback(((host, port), session_id, None), *callback_args)
+            callback = self.__switchboard_callbacks.pop(0)
+            run(callback, ((host, port), session_id, None))
 
     def _handle_USR(self, command):
         args_len = len(command.arguments)
@@ -351,7 +363,8 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
             if command.arguments[0] == "SSO":
                 self._client._sso.RequestMultipleSecurityTokens(
                     (self._sso_cb, command.arguments[3]),
-                    (lambda *args: self.emit("authentication-failed"),),
+                    ((lambda error: self.emit("error",
+                        ProtocolError.AUTHENTICATION_FAILED)),),
                     SSO.LiveService.MESSENGER_CLEAR)
 
                 self._client.address_book.connect("notify::state",
@@ -380,15 +393,19 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
     def _handle_SBS(self, command): # unknown command
         pass
 
+    def _handle_QNG(self, command):
+        timeout = int(command.arguments[0])
+        self.start_timeout("ping", timeout)
+
     def _handle_OUT(self, command):
         reason = None
         if len(command.arguments) > 0:
             reason = command.arguments[0]
 
         if reason == "OTH":
-            self.emit("disconnected-by-other")
+            self.emit("error", ProtocolError.OTHER_CLIENT)
         elif reason == "SSD":
-            self.emit("server-down")
+            self.emit("error", ProtocolError.SERVER_DOWN)
         else:
             self._transport.lose_connection()
 
@@ -397,6 +414,9 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         self._client.profile._server_property_changed("privacy",
                 command.arguments[0])
 
+    def _handle_GCF(self, command):
+        pass
+
     def _handle_CHG(self, command):
         self._client.profile._server_property_changed("presence",
                 command.arguments[0])
@@ -410,35 +430,19 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         else:
             self._client.profile._server_property_changed("msn_object", None)
 
-    def _handle_ILN(self,command):
+    def _handle_ILN(self, command):
         self._handle_NLN(command)
 
-    def _handle_FLN(self,command):
-        idx, network_id, account = self._parse_account(command)
-
-        contacts = self._client.address_book.contacts.\
-                search_by_network_id(network_id).\
-                search_by_account(account)
-
-        if len(contacts) == 0:
-            logger.warning("Contact (network_id=%d) %s not found" % \
-                    (network_id, account))
-
-        for contact in contacts:
+    def _handle_FLN(self, command):
+        idx, network_id, account = self.__parse_network_and_account(command)
+        contact = self.__search_account(account, network_id)
+        if contact is not None:
+            contact._remove_flag(profile.ContactFlag.EXTENDED_PRESENCE_KNOWN)
             contact._server_property_changed("presence",
                     profile.Presence.OFFLINE)
 
-    def _handle_NLN(self,command):
-        idx, network_id, account = self._parse_account(command, 1)
-
-        contacts = self._client.address_book.contacts.\
-                search_by_network_id(network_id).\
-                search_by_account(account)
-
-        if len(contacts) == 0:
-            logger.warning("Contact (network_id=%d) %s not found" % \
-                    (network_id, account))
-
+    def _handle_NLN(self, command):
+        idx, network_id, account = self.__parse_network_and_account(command, 1)
         presence = command.arguments[0]
         display_name = urllib.unquote(command.arguments[idx])
         idx += 1
@@ -447,19 +451,27 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
 
         msn_object = None
         icon_url = None
+
         if len(command.arguments) > idx:
             if command.arguments[idx] not in ('0', '1'):
                 msn_object = papyon.p2p.MSNObject.parse(self._client,
                                urllib.unquote(command.arguments[idx]))
+
         idx += 1
         if len(command.arguments) > idx:
             icon_url = command.arguments[idx]
 
-        for contact in contacts:
-            contact._server_property_changed("presence", presence)
+        contact = self.__search_account(account, network_id)
+        if contact is not None:
+            # don't change local presence and capabilities
+            if contact is not self._client.profile:
+                contact._server_property_changed("presence", presence)
+                contact._server_property_changed("client-capabilities", capabilities)
             contact._server_property_changed("display-name", display_name)
-            contact._server_property_changed("client-capabilities", capabilities)
-            contact._server_property_changed("msn-object", msn_object)
+            # only change MSNObject if the extended presence is known (MSNP18)
+            if self._protocol_version < 18 or \
+               contact.has_flag(profile.ContactFlag.EXTENDED_PRESENCE_KNOWN):
+                contact._server_property_changed("msn_object", msn_object)
             if icon_url is not None:
                 contact._server_attribute_changed('icon_url', icon_url)
 
@@ -475,17 +487,20 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
     def _handle_UUX(self, command):
         pass
 
-    def _handle_UBN(self,command): # contact infos
+    def _handle_UBN(self, command): # contact infos
         if not command.payload:
             return
+        account, guid = parse_account(command.arguments[0])
+        contact = self.__search_account(account)
         type = int(command.arguments[1])
-        self.emit("buddy-notification-received", type, command)
+        payload = command.payload
+        self.emit("buddy-notification-received", contact, guid, type, payload)
 
-    def _handle_UBX(self,command): # contact infos
+    def _handle_UBX(self, command): # contact infos
         if not command.payload:
             return
-
-        idx, network_id, account = self._parse_account(command)
+        print command.payload
+        idx, network_id, account = self.__parse_network_and_account(command)
 
         try:
             tree = ElementTree.fromstring(command.payload)
@@ -493,43 +508,48 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
             logger.error("Invalid XML data in received UBX command")
             return
 
-        cm = tree.find("./CurrentMedia")
-        if cm is not None and cm.text is not None:
-            parts = cm.text.split('\\0')
-            if parts[1] == 'Music' and parts[2] == '1':
-                cm = (parts[4].encode("utf-8"), parts[5].encode("utf-8"))
-            elif parts[2] == '0':
-                cm = None
-        else:
-            cm = None
+        utl = self.__find_node(tree, "./UserTileLocation", "")
+        cm_parts = self.__find_node(tree, "./CurrentMedia", "").split('\\0')
+        pm = self.__find_node(tree, "./PSM", "")
+        rmu = self.__find_node(tree, "./RMU", "")
+        ss = self.__find_node(tree, "./SignatureSound", None)
+        mg = self.__find_node(tree, "./MachineGuid", "{}").lower()[1:-1]
 
-        pm = tree.find("./PSM")
-        if pm is not None and pm.text is not None:
-            pm = pm.text.encode("utf-8")
-        else:
-            pm = ""
+        msn_object = None
+        if utl != "" and utl != "0":
+            msn_object = papyon.p2p.MSNObject.parse(self._client, utl)
 
-        ss = tree.find("./SignatureSound")
-        if ss is not None and ss.text is not None:
-            ss = ss.text.encode("utf-8")
-        else:
-            ss = None
+        cm = None
+        if len(cm_parts) >= 6 and cm_parts[1] == 'Music' and cm_parts[2] == '1':
+            cm = (cm_parts[4], cm_parts[5])
 
-        contacts = self._client.address_book.contacts.\
-                search_by_network_id(network_id).\
-                search_by_account(account)
+        eps = tree.findall("./EndpointData")
+        end_points = {}
+        for ep in eps:
+            guid = uuid.UUID(ep.get("id"))
+            caps = self.__find_node(ep, "Capabilities", "0:0")
+            end_points[guid] = profile.EndPoint(guid, caps)
+        peps = tree.findall("./PrivateEndpointData")
+        for pep in peps:
+            guid = uuid.UUID(pep.get("id"))
+            if guid not in end_points: continue
+            end_point = end_points[guid]
+            end_point.name = self.__find_node(pep, "EpName", "")
+            end_point.idle = bool(self.__find_node(pep, "Idle", "").lower() == "true")
+            end_point.client_type = int(self.__find_node(pep, "ClientType", 0))
+            end_point.state = self.__find_node(pep, "State", "")
 
-        if len(contacts) == 0:
-            logger.warning("Contact (network_id=%d) %s not found" % \
-                    (network_id, account))
-
-        for contact in contacts:
+        contact = self.__search_account(account, network_id)
+        if contact is not None:
+            contact._add_flag(profile.ContactFlag.EXTENDED_PRESENCE_KNOWN)
+            contact._server_property_changed("end-points", end_points)
+            contact._server_property_changed("msn-object", msn_object)
             contact._server_property_changed("current-media", cm)
             contact._server_property_changed("personal-message", pm)
             contact._server_property_changed("signature-sound", ss)
 
-    def _handle_UUN(self,command): # UBN acknowledgment
-        pass
+    def _handle_UUN(self, command): # UBN acknowledgment
+        self._command_answered_cb(command.transaction_id)
 
     # --------- Contact List -------------------------------------------------
     def _handle_ADL(self, command):
@@ -545,6 +565,9 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
     def _handle_RML(self, command):
         pass
 
+    def _handle_FQY(self, command):
+        pass
+
     # --------- Messages -----------------------------------------------------
     def _handle_MSG(self, command):
         message = Message(None, command.payload)
@@ -559,8 +582,7 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
                     profile[name] = value.strip()
             self._client.profile._server_property_changed("profile", profile)
 
-            self._send_command("BLP",
-                    (self._client.profile.privacy,))
+            self.set_privacy(self._client.profile.privacy)
             self._state = ProtocolState.SYNCHRONIZING
             self._client.address_book.sync()
         elif content_type[0] in \
@@ -590,9 +612,9 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
             #New mail
             m = HTTPMessage()
             m.parse(message.body)
-            name = m.get_header('From')
+            name = decode_rfc2047_string(m.get_header('From'))
             address = m.get_header('From-Addr')
-            subject = m.get_header('Subject')
+            subject = decode_rfc2047_string(m.get_header('Subject'))
             message_url = m.get_header('Message-URL')
             post_url = m.get_header('Post-URL')
             post_id = m.get_header('id')
@@ -616,17 +638,9 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
                 self._client.mailbox._unread_mail_increased(delta)
 
     def _handle_UBM(self, command):
-        idx, network_id, account = self._parse_account(command)
-
-        contacts = self._client.address_book.contacts.\
-                search_by_network_id(network_id).\
-                search_by_account(account)
-
-        if len(contacts) == 0:
-            logger.warning("Contact (network_id=%d) %s not found" % \
-                    (network_id, account))
-        else:
-            contact = contacts[0]
+        idx, network_id, account = self.__parse_network_and_account(command)
+        contact = self.__search_account(account, network_id)
+        if contact is not None:
             message = Message(contact, command.payload)
             self.emit("unmanaged-message-received", contact, message)
 
@@ -662,16 +676,14 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
 
     def _handle_URL(self, command):
         tr_id = command.transaction_id
-        if tr_id in self._url_callbacks:
+        if tr_id in self._callbacks:
             message_url, post_url, post_id = command.arguments
             post_url, form_dict = self._build_url_post_data(message_url,
                                                             post_url, post_id)
-            callback = self._url_callbacks[tr_id]
-            del self._url_callbacks[tr_id]
-            callback(post_url, form_dict)
+            self._command_answered_cb(tr_id, post_url, form_dict)
 
     # --------- Invitation ---------------------------------------------------
-    def _handle_RNG(self,command):
+    def _handle_RNG(self, command):
         session_id = command.arguments[0]
         host, port = command.arguments[1].split(':',1)
         port = int(port)
@@ -684,17 +696,27 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         self.emit("switchboard-invitation-received", session, inviter)
 
     # --------- Challenge ----------------------------------------------------
-    def _handle_QNG(self,command):
+    def _handle_QNG(self, command):
         pass
 
-    def _handle_QRY(self,command):
+    def _handle_QRY(self, command):
         pass
 
-    def _handle_CHL(self,command):
+    def _handle_CHL(self, command):
         response = _msn_challenge(command.arguments[0])
         self._send_command('QRY',
                 (ProtocolConstant.PRODUCT_ID,), payload=response)
 
+    # --------- Notification -------------------------------------------------
+    def _handle_NOT(self, command):
+        pass
+
+    #---------- Errors -------------------------------------------------------
+    def _error_handler(self, error):
+        logger.error('Notification got error : ' + unicode(error))
+        self._command_error_cb(error.transaction_id, int(error.name))
+
+
     # callbacks --------------------------------------------------------------
     def _connect_cb(self, transport):
         self.__switchboard_callbacks = PriorityQueue()
@@ -706,6 +728,7 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         self._send_command('VER', versions)
 
     def _disconnect_cb(self, transport, reason):
+        self.stop_all_timeout()
         self._state = ProtocolState.CLOSED
 
     def _sso_cb(self, tokens, nonce):
@@ -717,25 +740,44 @@ class NotificationProtocol(BaseProtocol, gobject.GObject):
         blob = clear_token.mbi_crypt(nonce)
         if self._protocol_version >= 16:
             arguments = ("SSO", "S", token, blob, "{%s}" %
-                    self._client.machine_guid.upper())
+                    str(self._client.machine_guid).upper())
         else:
             arguments = ("SSO", "S", token, blob)
 
         self._send_command("USR", arguments)
 
+    def on_ping_timeout(self):
+        self._transport.enable_ping()
+
+    def _command_answered_cb(self, tr_id, *args):
+        callback, errback = self._callbacks.get(tr_id, (None, None))
+        run(callback, *args)
+
+    def _command_error_cb(self, tr_id, error):
+        callback, errback = self._callbacks.get(tr_id, (None, None))
+        run(errback, error)
+
     def _address_book_state_changed_cb(self, address_book, pspec):
         MAX_PAYLOAD_SIZE = 7500
         if self._state != ProtocolState.SYNCHRONIZING:
             return
         if address_book.state != AB.AddressBookState.SYNCHRONIZED:
             return
+        self._client.profile._cid = address_book.profile._cid
         self._client.profile._server_property_changed("display-name",
                 address_book.profile.display_name)
 
         contacts = address_book.contacts.group_by_domain()
+        mask = ~(profile.Membership.REVERSE | profile.Membership.PENDING)
+
+        for contact in address_book.contacts:
+            if (contact.memberships & mask & ~profile.Membership.FORWARD) == \
+                    (profile.Membership.ALLOW | profile.Membership.BLOCK):
+                logger.warning("Contact is on both Allow and Block list; " \
+                               "removing from Allow list (%s)" % contact.account)
+                contact._remove_membership(profile.Membership.ALLOW)
 
         payloads = ['<ml l="1">']
-        mask = ~(profile.Membership.REVERSE | profile.Membership.PENDING)
         for domain, contacts in contacts.iteritems():
             payloads[-1] += '<d n="%s">' % domain
             for contact in contacts:
