@@ -5,6 +5,7 @@
 # Copyright (C) 2005-2006 Ali Sabil <ali.sabil@gmail.com>
 # Copyright (C) 2006  Johann Prieur <johann.prieur@gmail.com>
 # Copyright (C) 2006  Ole André Vadla Ravnås <oleavr@gmail.com>
+# Copyright (C) 2010  Collabora Ltd.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -30,6 +31,9 @@ inside HTTP POST requests.
 The classes of this module are structured as follow:
 G{classtree BaseTransport}"""
 
+from gnet.proxy.factory import ProxyFactory
+from papyon.util.async import run
+
 import gnet
 import gnet.protocol
 import msnp
@@ -145,8 +149,11 @@ class BaseTransport(gobject.GObject):
         """Connect to the server server"""
         raise NotImplementedError
 
-    def lose_connection(self):
-        """Disconnect from the server"""
+    def lose_connection(self, error=None):
+        """Disconnect from the server
+        
+            @param error: the error that caused the disconnection if any
+            @type error: Exception"""
         raise NotImplementedError
 
     def reset_connection(self, server=None):
@@ -158,7 +165,8 @@ class BaseTransport(gobject.GObject):
         raise NotImplementedError
 
     # Command Sending
-    def send_command(self, command, increment=True, callback=None, *cb_args):
+    def send_command(self, command, increment=True, callback=None,
+            errback=None):
         """
         Sends a L{msnp.Command} to the server.
 
@@ -168,17 +176,16 @@ class BaseTransport(gobject.GObject):
             @param increment: if False, the transaction ID is not incremented
             @type increment: bool
 
-            @param callback: callback to be used when the command has been
-                transmitted
-            @type callback: callable
+            @param callback: callback once transmission of command succeeded
+            @type callback: tuple (callable, args)
 
-            @param cb_args: callback arguments
-            @type cb_args: Any, ...
+            @param errback: callback if transmission of command failed
+            @type errback: tuple (callable, args)
         """
         raise NotImplementedError
 
-    def send_command_ex(self, command, arguments=(), payload=None, 
-            increment=True, callback=None, *cb_args):
+    def send_command_ex(self, command, arguments=(), payload=None,
+            increment=True, callback=None, errback=None):
         """
         Builds a command object then send it to the server.
         
@@ -195,16 +202,15 @@ class BaseTransport(gobject.GObject):
             @param increment: if False, the transaction ID is not incremented
             @type increment: bool
 
-            @param callback: callback to be used when the command has been
-                transmitted
-            @type callback: callable
+            @param callback: callback once transmission of command succeeded
+            @type callback: tuple (callable, args)
 
-            @param cb_args: callback arguments
-            @type cb_args: tuple
+            @param errback: callback if transmission of command failed
+            @type errback: tuple (callable, args)
         """
         cmd = msnp.Command()
         cmd.build(command, self._transaction_id, payload, *arguments)
-        self.send_command(cmd, increment, callback, *cb_args)
+        self.send_command(cmd, increment, callback, errback)
         return cmd
 
     def enable_ping(self):
@@ -225,23 +231,27 @@ class DirectConnection(BaseTransport):
     def __init__(self, server, server_type=ServerType.NOTIFICATION, proxies={}):
         BaseTransport.__init__(self, server, server_type, proxies)
 
-        transport = gnet.io.TCPClient(server[0], server[1])
-        transport.connect("notify::status", self.__on_status_change)
-        transport.connect("error", self.__on_error)
-
+        transport = self._setup_transport(server[0], server[1], proxies)
         receiver = gnet.parser.DelimiterParser(transport)
         receiver.connect("received", self.__on_received)
 
         self._receiver = receiver
         self._receiver.delimiter = "\r\n"
         self._transport = transport
-        self.__pending_chunk = None
+        self.__pending_command = None
         self.__resetting = False
-        self.__error = False
-        self.__png_timeout = None
+        self.__error = None
 
     __init__.__doc__ = BaseTransport.__init__.__doc__
 
+    def _setup_transport(self, host, port, proxies):
+        transport = gnet.io.TCPClient(host, port)
+        if proxies:
+            transport = ProxyFactory(transport, proxies, 'direct')
+        transport.connect("notify::status", self.__on_status_change)
+        transport.connect("error", self.__on_error)
+        return transport
+
     ### public commands
 
     @property
@@ -249,15 +259,17 @@ class DirectConnection(BaseTransport):
         return self._transport.sockname
 
     def establish_connection(self):
+        self.__pending_command = None
         logger.debug('<-> Connecting to %s:%d' % self.server )
         self._transport.open()
 
-    def lose_connection(self):
+    def lose_connection(self, error=None):
+        self.__error = error
+        if error is not None:
+            self.emit("connection-failure", error)
+
         self.__resetting = False
         self._transport.close()
-        if self.__png_timeout is not None:
-            gobject.source_remove(self.__png_timeout)
-            self.__png_timeout = None
 
     def reset_connection(self, server=None):
         if server:
@@ -265,16 +277,18 @@ class DirectConnection(BaseTransport):
             self._transport.set_property("port", server[1])
             self.server = server
         self.__resetting = True
-        if self.__png_timeout is not None:
-            gobject.source_remove(self.__png_timeout)
-            self.__png_timeout = None
         self._transport.close()
         self._transport.open()
 
-    def send_command(self, command, increment=True, callback=None, *cb_args):
+    def send_command(self, command, increment=True, callback=None,
+            errback=None):
+        if self.__error:
+            logger.warning("Transport is errored, not sending %s" % command.name)
+            run(errback, self.__error)
+            return
         logger.debug('>>> ' + unicode(command))
-        our_cb_args = (command, callback, cb_args)
-        self._transport.send(str(command), self.__on_command_sent, *our_cb_args)
+        self._transport.send(str(command),
+                (self.__on_command_sent, command, callback), errback)
         if increment:
             self._increment_transaction_id()
 
@@ -282,17 +296,10 @@ class DirectConnection(BaseTransport):
         cmd = msnp.Command()
         cmd.build("PNG", None)
         self.send_command(cmd, False)
-        self.__png_timeout = None
-        return False
 
-    def __on_command_sent(self, command, user_callback, user_cb_args):
+    def __on_command_sent(self, command, user_callback):
         self.emit("command-sent", command)
-        if user_callback:
-            user_callback(*user_cb_args)
-
-    def __handle_ping_reply(self, command):
-        timeout = int(command.arguments[0])
-        self.__png_timeout = gobject.timeout_add(timeout * 1000, self.enable_ping)
+        run(user_callback)
 
     ### callbacks
     def __on_status_change(self, transport, param):
@@ -305,40 +312,36 @@ class DirectConnection(BaseTransport):
         elif status == gnet.IoStatus.CLOSED:
             if not self.__resetting and not self.__error:
                 self.emit("connection-lost", None)
-            self.__error = False
+            self.__error = None
     
-    def __on_error(self, transport, reason):
+    def __on_error(self, transport, error):
         status = transport.get_property("status")
-        self.__error = True
+        self.__error = error
         if status == gnet.IoStatus.OPEN:
-            self.emit("connection-lost", reason)
+            self.emit("connection-lost", error)
         else:
-            self.emit("connection-failure", reason)
+            self.emit("connection-failure", error)
 
     def __on_received(self, receiver, chunk):
-        cmd = msnp.Command()
-        if self.__pending_chunk:
-            chunk = self.__pending_chunk + "\r\n" + chunk
-            cmd.parse(chunk)
-            self.__pending_chunk = None
+        if self.__pending_command is None:
+            cmd = msnp.Command()
+            try:
+                cmd.parse(chunk)
+            except Exception, err:
+                logger.error("Received invalid command, closing connection")
+                self.lose_connection(err)
+                return
+            if cmd.payload_len > 0:
+                self.__pending_command = cmd
+                self._receiver.delimiter = cmd.payload_len
+                return # wait for payload
+        else:
+            cmd = self.__pending_command
+            cmd.payload = chunk
+            self.__pending_command = None
             self._receiver.delimiter = "\r\n"
-        else:
-            cmd.parse(chunk)
-            if cmd.name in msnp.Command.INCOMING_PAYLOAD or \
-                    (cmd.is_error() and (cmd.arguments is not None) and len(cmd.arguments) > 0):
-                try:
-                    payload_len = int(cmd.arguments[-1])
-                except:
-                    payload_len = 0
-                if payload_len > 0:
-                    self.__pending_chunk = chunk
-                    self._receiver.delimiter = payload_len
-                    return
         logger.debug('<<< ' + unicode(cmd))
-        if cmd.name == 'QNG':
-            self.__handle_ping_reply(cmd)
-        else:
-            self.emit("command-received", cmd)
+        self.emit("command-received", cmd)
 gobject.type_register(DirectConnection)
 
 
@@ -350,44 +353,54 @@ class HTTPPollConnection(BaseTransport):
         self._target_server = server
         server = ("gateway.messenger.hotmail.com", 80)
         BaseTransport.__init__(self, server, server_type, proxies)
-        self._setup_transport()
+        self._setup_transport(server[0], server[1], proxies)
         
         self._command_queue = []
         self._waiting_for_response = False # are we waiting for a response
         self._session_id = None
-        self.__error = False
+        self.__error = None
 
-    def _setup_transport(self):
-        server = self.server
-        proxies = self.proxies
-        if 'http' in proxies:
-            transport = gnet.protocol.HTTP(server[0], server[1], proxies['http'])
-        else:
-            transport = gnet.protocol.HTTP(server[0], server[1])
-        transport.connect("response-received", self.__on_received)
-        transport.connect("request-sent", self.__on_sent)
-        transport.connect("error", self.__on_error)
+    def _setup_transport(self, host, port, proxies):
+        handles = []
+        transport = gnet.protocol.HTTP(host, port, proxies)
+        handles.append(transport.connect("response-received", self.__on_received))
+        handles.append(transport.connect("request-sent", self.__on_sent))
+        handles.append(transport.connect("error", self.__on_error))
+        self._transport_handles = handles
         self._transport = transport
 
     def establish_connection(self):
         logger.debug('<-> Connecting to %s:%d' % self.server)
-        self._polling_source_id = gobject.timeout_add(5000, self._poll)
+        self._polling_source_id = gobject.timeout_add_seconds(5, self._poll)
         self.emit("connection-success")
 
-    def lose_connection(self):
+    def lose_connection(self, error=None):
         gobject.source_remove(self._polling_source_id)
         del self._polling_source_id
-        if not self.__error:
+        if error is not None:
+            self.emit("connection-failure", error)
+        elif not self.__error:
             self.emit("connection-lost", None)
-        self.__error = False
+        self.__error = None
 
     def reset_connection(self, server=None):
         if server:
             self._target_server = server
         self.emit("connection-reset")
 
-    def send_command(self, command, increment=True, callback=None, *cb_args):
-        self._command_queue.append((command, increment, callback, cb_args))
+    def change_gateway(self, server):
+        if self.server == server:
+            return
+        logger.debug('<-> Changing gateway to %s:%d' % server)
+        self.server = server
+        for handle in self._transport_handles:
+            self._transport.disconnect(handle)
+        self._transport.close()
+        self._setup_transport(server[0], server[1], self.proxies)
+
+    def send_command(self, command, increment=True, callback=None,
+            errback=None):
+        self._command_queue.append((command, increment, callback, errback))
         self._send_command()
 
     def _send_command(self):
@@ -429,14 +442,12 @@ class HTTPPollConnection(BaseTransport):
             self.send_command(None)
         return True
     
-    def __on_error(self, transport, reason):
-        self.__error = True
-        self.emit("connection-lost", reason)
+    def __on_error(self, transport, error):
+        self.__error = error
+        self.emit("connection-lost", error)
+        self.lose_connection()
         
     def __on_received(self, transport, http_response):
-        if http_response.status == 403:
-            self.emit("connection-lost", TransportError.PROXY_FORBIDDEN)
-            self.lose_connection()
         if 'X-MSN-Messenger' in http_response.headers:
             data = http_response.headers['X-MSN-Messenger'].split(";")
             for elem in data:
@@ -444,8 +455,7 @@ class HTTPPollConnection(BaseTransport):
                 if key == 'SessionID':
                     self._session_id = value
                 elif key == 'GW-IP':
-                    self.server = (value, self.server[1])
-                    self._setup_transport()
+                    self.change_gateway((value, self.server[1]))
                 elif key == 'Session'and value == 'close':
                     #self.lose_connection()
                     pass
@@ -461,31 +471,30 @@ class HTTPPollConnection(BaseTransport):
     def __on_sent(self, transport, http_request):
         if len(self._command_queue) == 0:
             return
-        command, increment, callback, cb_args = self._command_queue.pop(0)
+        command, increment, callback, errback = self._command_queue.pop(0)
         if command is not None:
-            if callback:
-                callback(*cb_args)
+            run(callback)
             self.emit("command-sent", command)
 
     def __extract_command(self, data):
-        first, rest = data.split('\r\n', 1)
-        cmd = msnp.Command()
-        cmd.parse(first.strip())
-        if cmd.name in msnp.Command.INCOMING_PAYLOAD or \
-                (cmd.is_error() and (cmd.arguments is not None) and len(cmd.arguments) > 0):
-            try:
-                payload_len = int(cmd.arguments[-1])
-            except:
-                payload_len = 0
-            if payload_len > 0:
-                cmd.payload = rest[:payload_len].strip()
-            logger.debug('<<< ' + unicode(cmd))
-            self.emit("command-received", cmd)
-            return rest[payload_len:]
-        else:
-            logger.debug('<<< ' + unicode(cmd))
-            self.emit("command-received", cmd)
-            return rest
-
+        try:
+            first, rest = data.split('\r\n', 1)
+        except ValueError, err:
+            logger.warning('Unable to extract a command: %s' % data)
+            self.lose_connection(err)
+            return []
 
+        try:
+            cmd = msnp.Command()
+            cmd.parse(first.strip())
+            if cmd.payload_len > 0:
+                cmd.payload = rest[:cmd.payload_len].strip()
+                rest = rest[cmd.payload_len:]
+        except Exception, err:
+            logger.error("Received invalid command, closing connection")
+            self.lose_connection(err)
+            return []
 
+        logger.debug('<<< ' + unicode(cmd))
+        self.emit("command-received", cmd)
+        return rest
